/***************************************************************************
 *   Copyright (C) 2005-2006 by Georg Hennig                               *
 *   Email: georg.hennig@web.de                                            *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include <qbuttongroup.h>
#include <qcheckbox.h>
#include <qclipboard.h>
#include <qdir.h>
#include <qlabel.h>
#include <qradiobutton.h>
#include <qtabwidget.h>
#include <qobject.h>

#include <kaboutapplication.h>
#include <kcombobox.h>
#include <kdebug.h>
#include <khelpmenu.h>
#include <kiconloader.h>
#include <kio/netaccess.h>
#include <kled.h>
#include <klineedit.h>
#include <klistview.h>
#include <kmessagebox.h>
#include <knuminput.h>
#include <kpopupmenu.h>
#include <kprogress.h>
#include <kpropertiesdialog.h>
#include <kpushbutton.h>
#include <krun.h>
#include <ktempfile.h>
#include <kurldrag.h>
#include <kurlrequester.h>

#include "kdatecombo.h"
#include "kquery.h"
#include "komparatorwidget.h"
#include "komparejob.h"
#include "kfileitemext.h"
#include "klistviewitemsingle.h"
#include "klistviewitemdups.h"

#include "config.h"

KomparatorWidget::KomparatorWidget( QWidget* parent, const char* name, WFlags fl )
        : KomparatorWidgetBase( parent, name, fl )
{
  url1_KURLRequester->setFocus();

  kompare = new KompareJob( this );

  query = new KQuery( this );

  KHelpMenu *help_menu = new KHelpMenu( this, "" );
  connect( help_KPushButton, SIGNAL( clicked() ), help_menu, SLOT( appHelpActivated() ) );

  connect( query, SIGNAL( addFile( const KFileItem*, const QString& ) ),
           SLOT( slotAddFile( const KFileItem*, const QString& ) ) );

  connect( query, SIGNAL( result( int, QString ) ), SLOT( slotResult( int, QString ) ) );

  connect( duplicates1_KListView, SIGNAL( selectionChanged() ), SLOT( slotDuplicateSelected() ) );

  connect( this, SIGNAL( signalStatus( QString ) ), SLOT( slotStatus( QString ) ) );
  connect( this, SIGNAL( signalProgress( QString, int ) ), SLOT( slotProgress( QString, int ) ) );

  connect( this, SIGNAL( signalSaveListViewColumnWidth() ), SLOT( slotSaveListViewColumnWidth() ) );

  files_dupes = NULL;
  files_comp = NULL;
  dirs_dupes = NULL;

  popupMenu = NULL;

  requested_download = NULL;
  requested_download_job = NULL;

  status = CLEAN;

  unreadable_files = 0;

  duplicatesTrash_KPushButton->setIconSet( DesktopIcon( "edittrash" ) );
  duplicatesTrash_KPushButton->setText( "" );

  duplicatesDelete_KPushButton->setIconSet( DesktopIcon( "editdelete" ) );
  duplicatesDelete_KPushButton->setText( "" );

  scan_KPushButton->setIconSet( SmallIcon( "komparator" ) );

  about_KPushButton->setIconSet( SmallIcon( "komparator" ) );
  help_KPushButton->setIconSet( SmallIcon( "help" ) );

  emit( signalSaveListViewColumnWidth() );
}

KomparatorWidget::~KomparatorWidget()
{
  query->kill();

  kompare->stop();

  if ( requested_download ) // Killing current download job and md5sum /
  {                         // byte array operation. Temporary file will be deleted.
    requested_download->setDownloadError( KIO::ERR_USER_CANCELED );
  }
  requested_download = NULL;

  if ( requested_download_job )
  {
    KIO::NetAccess::del( ((KIO::FileCopyJob*)requested_download_job)->destURL().url(), this );
    KIO::NetAccess::del( ((KIO::FileCopyJob*)requested_download_job)->destURL().url()+".part", this ); // FIXME: ugly.
    requested_download_job->kill();
  }
  requested_download_job = NULL;

  kompare->wait(); // kompare uses files_dupes etc., so we need to wait until kompare has really finished.

  delete query;
  delete kompare;

  duplicates1_KListView->clear();
  duplicates2_KListView->clear();

  missing1_KListView->clear();
  missing2_KListView->clear();

  newer1_KListView->clear();
  newer2_KListView->clear();

  if ( files_dupes )
  {
    delete files_dupes;
    files_dupes = NULL;
  }

  if ( files_comp )
  {
    delete files_comp;
    files_comp = NULL;
  }

  if ( dirs_dupes )
  {
    delete dirs_dupes;
    dirs_dupes = NULL;
  }

  if ( popupMenu )
  {
    delete popupMenu;
    popupMenu = NULL;
  }
}



/****************************************
 * Drag & Drop Functions                *
 ****************************************/

QDragObject *KListView::dragObject()
{
  KURL::List urls;

  QPixmap *pm = NULL;

  QListViewItemIterator it( this );
  while ( it.current() )
  {
    if ( it.current()->isSelected() )
    {
      urls.append( ((KListViewItemSingle*)it.current())->item->url().url() );
      if ( pm == NULL )
      {
        pm = new QPixmap( ((KListViewItemSingle*)it.current())->item->pixmap( 32 ) );
      }
    }
    it++;
  }

  KURLDrag *ret = new KURLDrag( urls, this );

  if ( pm != NULL && !pm->isNull() ) ret->setPixmap( *pm );

  return ret;
}

void KomparatorWidget::slotKListViewDropped( KListView *list_view, QDropEvent *, QListViewItem * )
{
  if ( list_view == missing1_KListView ) currentListView = missing2_KListView; // set origin of drag.
  else if ( list_view == missing2_KListView ) currentListView = missing1_KListView;
  else if ( list_view == newer1_KListView ) currentListView = newer2_KListView;
  else if ( list_view == newer2_KListView ) currentListView = newer1_KListView;

  slotCopyToOtherSide();
}

bool KListView::acceptDrag ( QDropEvent *event ) const // acceptDrag only under certain circumstances:
{
  if ( !event->source() || !this ) return FALSE; // no acceptable source (only missing1 & 2 & newer1 & 2).

  if ( ( qstrcmp( event->source()->name(), "missing1_KListView" ) == 0
         && qstrcmp( this->name(), "missing2_KListView" ) == 0  ) ||
       ( qstrcmp( event->source()->name(), "missing2_KListView" ) == 0
         && qstrcmp( this->name(), "missing1_KListView" ) == 0  ) ||
       ( qstrcmp( event->source()->name(), "newer1_KListView" ) == 0
         && qstrcmp( this->name(), "newer2_KListView" ) == 0  ) ||
       ( qstrcmp( event->source()->name(), "newer2_KListView" ) == 0
         && qstrcmp( this->name(), "newer1_KListView" ) == 0  ) )
  {
    return TRUE;
  }

  return FALSE;
}



/****************************************
 * Keyboard Functions                   *
 ****************************************/

void KListView::keyPressEvent( QKeyEvent *event )
{
  switch( event->key() ) // ignoring events passes them to the parent (=KomparatorWidget).
  {
    case Qt::Key_Delete:
      event->ignore();
      return;
    case Qt::Key_C:
      if ( event->state() == Qt::ControlButton )
      {
        event->ignore();
        return;
      }
      break;
    case Qt::Key_Enter:  // Enter and return are equal events for us.
    case Qt::Key_Return:
      event->ignore();
      return;
    default:
      break;
  }

  event->accept();
  QListView::keyPressEvent( event );
}

void KomparatorWidget::keyPressEvent( QKeyEvent *event )
{
  if ( duplicates1_KListView->hasFocus() ) currentListView = duplicates1_KListView;
  else if ( duplicates2_KListView->hasFocus() ) currentListView = duplicates2_KListView;
  else if ( missing1_KListView->hasFocus() ) currentListView = missing1_KListView;
  else if ( missing2_KListView->hasFocus() ) currentListView = missing2_KListView;
  else if ( newer1_KListView->hasFocus() ) currentListView = newer1_KListView;
  else if ( newer2_KListView->hasFocus() ) currentListView = newer2_KListView;
  else
  {
    event->ignore();
    return;
  }

  switch( event->key() ) // ignoring events passes them to the parent (=KomparatorWidget).
  {
    case Qt::Key_Delete:
      if ( event->state() == Qt::ShiftButton )
      {
        slotDeleteFiles();
      }
      else
      {
        slotTrashFiles();
      }
      event->accept();
      return;
    case Qt::Key_C:
      if ( event->state() == Qt::ControlButton )
      {
        event->accept();
        slotCopySelection();
        return;
      }
      break;
    case Qt::Key_Enter:  // Enter and return are equal events for us.
    case Qt::Key_Return:
      event->accept();
      slotOpenBinding();
      return;
    default:
      break;
  }

  event->ignore();
}



/****************************************
 * Context Menu Functions               *
 ****************************************/

void KomparatorWidget::slotContextMenu( KListView *list_view, QListViewItem *item, const QPoint &p )
{
  if ( item == NULL ) return;
  if ( list_view == NULL ) return;

  int count = list_view->selectedItems().count();

  currentListView = list_view;

  if ( count == 0 ) return;

  if ( popupMenu == NULL )
     popupMenu = new KPopupMenu( this );
  else
     popupMenu->clear();

  if ( count == 1 )
  {
    QString title;
    if ( currentListView == duplicates1_KListView ) title = item->text( 0 )+item->text( 1 );
    else title = item->text( 1 )+item->text( 0 );
    popupMenu->insertTitle( ((KListViewItemSingle *)item)->item->pixmap( 16 ), title );
    popupMenu->insertItem( SmallIcon( "editdelete" ), i18n( "Remove from list" ),
                           this, SLOT( slotRemoveFromList() ) );
    if ( list_view == missing1_KListView || list_view == newer1_KListView )
      popupMenu->insertItem( SmallIcon( "editcopy" ), i18n( "Copy to right side" ),
                             this, SLOT( slotCopyToOtherSide() ) );
    else if ( list_view == missing2_KListView || list_view == newer2_KListView )
      popupMenu->insertItem( SmallIcon( "editcopy" ), i18n( "Copy to left side" ),
                             this, SLOT( slotCopyToOtherSide() ) );
    popupMenu->insertSeparator();
    popupMenu->insertItem( SmallIcon( "fileopen" ), i18n( "Open" ),
                           this, SLOT( slotOpenBinding() ) );
    popupMenu->insertItem( SmallIcon( "window_new" ), i18n( "Open Folder" ),
                           this, SLOT( slotOpenFolder() ) );
    popupMenu->insertSeparator();
    popupMenu->insertItem( SmallIcon( "editcopy" ), i18n( "Copy" ),
                           this, SLOT( slotCopySelection() ) );
    popupMenu->insertItem( SmallIcon( "edittrash" ), i18n( "Trash" ),
                           this, SLOT( slotTrashFiles() ) );
    popupMenu->insertItem( SmallIcon( "editdelete" ), i18n( "Delete" ),
                           this, SLOT( slotDeleteFiles() ) );
    popupMenu->insertSeparator();
    popupMenu->insertItem( i18n( "Open With..." ),
                           this, SLOT( slotOpenWith() ) );
    popupMenu->insertSeparator();
    popupMenu->insertItem( i18n( "Properties" ),
                           this, SLOT( slotFileProperties() ) );
  }
  else
  {
    popupMenu->insertTitle( i18n( "Selected Files" ) );
    popupMenu->insertItem( SmallIcon( "editdelete" ), i18n( "Remove from list" ),
                           this, SLOT( slotRemoveFromList() ) );
    if ( list_view == missing1_KListView || list_view == newer1_KListView )
      popupMenu->insertItem( SmallIcon( "editcopy" ), i18n( "Copy to right side" ),
                             this, SLOT( slotCopyToOtherSide() ) );
    else if ( list_view == missing2_KListView || list_view == newer2_KListView )
      popupMenu->insertItem( SmallIcon( "editcopy" ), i18n( "Copy to left side" ),
                             this, SLOT( slotCopyToOtherSide() ) );
    popupMenu->insertSeparator();
    popupMenu->insertItem( SmallIcon( "editcopy" ), i18n( "Copy" ),
                           this, SLOT( slotCopySelection() ) );
    popupMenu->insertItem( SmallIcon( "edittrash" ), i18n( "Trash" ),
                           this, SLOT( slotTrashFiles() ) );
    popupMenu->insertItem( SmallIcon("editdelete"), i18n( "Delete" ),
                           this, SLOT( slotDeleteFiles() ) );
  }

  popupMenu->popup( p, 1 );
}



/****************************************
 * Interface Functions                  *
 ****************************************/

void KomparatorWidget::slotSearch()
{
  switch ( status )
  {
    case CLEAN:
      if ( !initializeQuery() ) return;

      scan1_KLed->setColor( QColor( 255, 0, 0 ) );  // Red
      scan2_KLed->setColor( QColor( 255, 0, 0 ) );
      scan3_KLed->setColor( QColor( 255, 0, 0 ) );

      enable( FALSE );

      unreadable_files = 0;
      number_of_files = 0;

      status = SEARCHING_URL_1;
      scan_KPushButton->setText( i18n( "Cancel" ) );
      scan_KPushButton->setIconSet( SmallIcon( "cancel" ) );

      emit( signalStatus( i18n( "Searching left URL for files..." ) ) );

      query->start();
      return;

    case SEARCHING_URL_1:
    case SEARCHING_URL_2:
      query->kill();
      break;

    case COMPARING:
      kompare->stop();

      if ( requested_download ) // Killing current download job and md5sum /
      {                         // byte array operation. Temporary file will be deleted.
        requested_download->setDownloadError( KIO::ERR_USER_CANCELED );
      }
      requested_download = NULL;

      if ( requested_download_job )
      {
        KIO::NetAccess::del( ((KIO::FileCopyJob*)requested_download_job)->destURL().url(), this );
        KIO::NetAccess::del( ((KIO::FileCopyJob*)requested_download_job)->destURL().url() + ".part", this ); // FIXME: ugly.
        requested_download_job->kill();
      }
      requested_download_job = NULL;

      kompare->wait(); // kompare uses files_dupes etc., so we need to wait until kompare has really finished.

      break;

    case READY:
      break;

    default:
      break;
  }

  duplicates1_KListView->clear();
  duplicates2_KListView->clear();

  missing1_KListView->clear();
  missing2_KListView->clear();

  newer1_KListView->clear();
  newer2_KListView->clear();

  if ( files_dupes )
  {
    delete files_dupes;
    files_dupes = NULL;
  }

  if ( files_comp )
  {
    delete files_comp;
    files_comp = NULL;
  }

  while ( dirs_dupes )
  {
    delete dirs_dupes;
    dirs_dupes = NULL;
  }

  scan1_KLed->setColor( QColor( 0, 255, 0 ) );  // Green
  scan2_KLed->setColor( QColor( 0, 255, 0 ) );
  scan3_KLed->setColor( QColor( 0, 255, 0 ) );

  enable( TRUE );

  unreadable_files = 0;
  number_of_files = 0;

  status = CLEAN;
  scan_KPushButton->setText( i18n( "Scan" ) );
  scan_KPushButton->setIconSet( SmallIcon( "komparator" ) );

  resizeListViewColumns();

  emit( signalStatus( i18n( "Ready." ) ) );
  emit( signalProgress( "", -1 ) );
}

void KomparatorWidget::slotAbout()
{
  KAboutApplication *about = new KAboutApplication( this );
  about->show();
}

void KomparatorWidget::slotDuplicateSelected()
{
  duplicates2_KListView->clear();

  if ( !duplicates1_KListView->selectedItem() ) return;

  KFileItemExt *tmpfile;

  KListViewItemSingle *tmpitem = new KListViewItemSingle( duplicates2_KListView,
                                       (((KListViewItemDups*)duplicates1_KListView->selectedItem())->item), NULL );
  tmpitem->setText( 1, tmpitem->item->url().prettyURL().replace( tmpitem->item->url().fileName(), "" ) );

  tmpfile = ((KListViewItemDups*)duplicates1_KListView->selectedItem())->item->duplicates;

  if ( !tmpfile ) return;

  if ( (((KListViewItemDups*)duplicates1_KListView->selectedItem())->item)->dir == 0 ) tmpitem->setText( 4, "L" );
  else tmpitem->setText( 4, "R" );

  while ( tmpfile != NULL )
  {
    tmpitem = new KListViewItemSingle( duplicates2_KListView, tmpfile, NULL );
    tmpitem->setText( 1, tmpitem->item->url().prettyURL().replace( tmpitem->item->url().fileName(), "" ) );

    if ( tmpfile->dir == 0 ) tmpitem->setText( 4, "L" );
    else tmpitem->setText( 4, "R" );

    tmpfile = tmpfile->duplicates;
  }

  duplicates2_KListView->setSorting( 4, !duplicatesReference1_RadioButton->isChecked() );

  duplicates2_KListView->setSelected( duplicates2_KListView->firstChild(), TRUE );
}

void KomparatorWidget::slotReferenceLeft( bool is_ref )
{
  duplicates2_KListView->setSorting(4, !is_ref );

  QListViewItemIterator it( duplicates2_KListView );
  while ( it.current() )
  {
    if ( it.current()->isSelected() )
    {
      it.current()->setSelected( FALSE );
    }
    it++;
  }

  duplicates2_KListView->setSelected( duplicates2_KListView->firstChild(), TRUE );
}

void KomparatorWidget::slotRemoveFromList()
{
  if ( currentListView == NULL ) return;

  emit( signalProgress( i18n( "Removing %1 files from list..." ).arg( currentListView->selectedItems().count() ), -1 ) );

  if ( currentListView == duplicates2_KListView ) // in all list views, the items are just removed. but here we'll need
                                                  // to remove them from the parent's (duplicates1_KListView) item (KListViewItemDups).
  {
    KFileItemExt *curfile;

    QListViewItemIterator it( duplicates2_KListView );
    while ( it.current() )
    {
      if ( it.current()->isSelected() )
      {
        curfile = ((KListViewItemDups*)duplicates1_KListView->selectedItem())->item;

        if ( curfile )
        {
          if ( ((KListViewItemSingle*)it.current())->item->url() == curfile->url() )
          {
            ((KListViewItemDups*)duplicates1_KListView->selectedItem())->item = curfile->duplicates;
            if ( curfile->duplicates ) curfile->duplicates = curfile->duplicates->duplicates; // Skip curfile->duplicates in duplicate chain.
          }

          while ( curfile->duplicates )
          {
            if ( ((KListViewItemSingle*)it.current())->item->url() == curfile->duplicates->url() )
            {
              curfile->duplicates = curfile->duplicates->duplicates;                          // Skip curfile->duplicates in duplicate chain.
            }
            if ( curfile->duplicates ) curfile = curfile->duplicates;
          }
        }
      }
      it++;
    }
  }

  currentListView->selectedItems().setAutoDelete( TRUE );

  if ( currentListView == duplicates2_KListView )
  {
    if ( !((KListViewItemDups*)duplicates1_KListView->selectedItem())->item )
    {
      duplicates1_KListView->selectedItems().setAutoDelete( TRUE );
    }
  }

  currentListView = NULL;

  emit( signalProgress( "", -1 ) );
}

void KomparatorWidget::slotCopyToOtherSide()
{
  if ( currentListView == NULL ) return;

  if ( currentListView->selectedItems().count() == 0 ) return;

  QString progress_str = i18n( "Copying %1 files from one side to the other..." ).arg( currentListView->selectedItems().count() );

  emit( signalProgress( progress_str, -1 ) );

  enable( FALSE );

  bool overwrite = FALSE; // doesn't make sense to overwrite if missing (then the target file shouldn't exist), but should overwrite if newer.
  if ( currentListView == newer1_KListView || currentListView == newer2_KListView )
  {
    QString tmp = i18n( "Do you really want to copy the %1 selected file overwriting the existing one?" ).arg( currentListView->selectedItems().count() );
    if ( KMessageBox::warningContinueCancel( parentWidget(), tmp, "", KGuiItem( i18n( "Overwrite" ), "editoverwrite") ) == KMessageBox::Cancel )
      return;
    overwrite = TRUE;
  }

  bool success;

  QString destination, source;
  if ( currentListView == missing1_KListView || currentListView == newer1_KListView ) // destination is right side
  {
    destination = url2_KURLRequester->url();
    source = url1_KURLRequester->url();
  }
  else if ( currentListView == missing2_KListView || currentListView == newer2_KListView ) // destination is left side
  {
    destination = url1_KURLRequester->url();
    source = url2_KURLRequester->url();
  }
  else // this point should never be reached.
  {
    kdDebug() << "Tried to copy from one to another side, but the event was not received from a valid list view." << endl;
    return;
  }

  int progress = 0;

  QListViewItemIterator it( currentListView );
  while ( it.current() )
  {
    if ( it.current()->isSelected() )
    {
      emit( signalProgress( progress_str, (int)((progress*100)/currentListView->selectedItems().count()) ) );

      KURL src( source );
      src.addPath( it.current()->text( 1 )+((KListViewItemSingle*)it.current())->item->url().fileName() );

      KURL dest( destination );
      dest.addPath( it.current()->text( 1 )+((KListViewItemSingle*)it.current())->item->url().fileName() );

      KURL dir( dest.directory( FALSE ) );

      while ( !KIO::NetAccess::mkdir( dir, this ) ) // create all destination subdirs if not present
      {
        KURL mkdir = dir.upURL();
        success = TRUE;

        while ( !KIO::NetAccess::mkdir( mkdir, this ) )
        {
          if ( mkdir == mkdir.upURL() )
          {
            success = FALSE;
            break;
          }

          mkdir = mkdir.upURL();
        }

        if ( !success ) break;
      }

      src.cleanPath();
      dest.cleanPath();

      if ( KIO::NetAccess::file_copy( src, dest, -1, overwrite ) ) // Try to copy. If subdirs were not created (not writeable, ...) it will fail.
      {
        // create a new item that holds the file that was just copied. 
        KFileItemExt *newitem;
        newitem = new KFileItemExt( dest, "", 0 );
// FIXME: on remote fs this gives 0 time and 0 size. see new empty folder, too.
        if ( currentListView == missing1_KListView || currentListView == newer1_KListView ) newitem->dir = 1;
        else newitem->dir = 0;
        slotAddFile( newitem, "" ); // Add it to files_dupes / files_comp to delete if necessary
        newitem->isdupe = 1;

        if ( !((KListViewItemSingle*)it.current())->item->brother->isdupe &&
             !((KListViewItemSingle*)it.current())->item->brother->hasdupes ) // no duplicate, copying created a new duplicate
        {
          newitem->duplicates = ((KListViewItemSingle*)it.current())->item->brother->duplicates;
          if ( newitem->duplicates != NULL ) newitem->hasdupes = 1;
          ((KListViewItemSingle*)it.current())->item->brother->duplicates = newitem;
          new KListViewItemDups( duplicates1_KListView, ((KListViewItemSingle*)it.current())->item->brother );
        }
        else // is already a duplicate, append copied file. must delete brother from duplicate chain.
        {
          KFileItemExt *curfile;
          curfile = ((KListViewItemSingle*)it.current())->item->brother->duplicates;
          ((KListViewItemSingle*)it.current())->item->brother->duplicates = newitem;
          ((KListViewItemSingle*)it.current())->item->brother->duplicates->duplicates = curfile;
        }

        if ( ((KListViewItemSingle*)it.current())->brother != NULL )
        {
          if ( ((KListViewItemSingle*)it.current())->brother->item->brother->dup_parent != NULL )
          {
            ((KListViewItemSingle*)it.current())->brother->item->brother->dup_parent->duplicates =
              ((KListViewItemSingle*)it.current())->brother->item->brother->duplicates;

            QListViewItemIterator it2( duplicates1_KListView );
            while( it2.current() )
            {
              if ( ((KListViewItemDups*)it2.current())->item->duplicates == NULL )
              {
                delete ((KListViewItemDups*)it2.current());
              }

              it2++;
            }
          }
        }

        delete ((KListViewItemSingle*)it.current())->brother; // remove the brother from the other list as it was overwritten.
      }
      else
      {
        KMessageBox::sorry( this, i18n( "Unable to copy %1." ).arg( src.url() + i18n( " to " ) + dest.url() ) );
        it.current()->setSelected( FALSE );
      }

      progress++;
    }
    it++;
  }

  emit( signalProgress( progress_str, 100 ) );

  currentListView->selectedItems().setAutoDelete( TRUE );

  enable( TRUE );

  currentListView = NULL;

  emit( signalProgress( "", -1 ) );
}

void KomparatorWidget::slotOpenBinding()
{
  if ( currentListView != NULL )
  {
    ((KListViewItemSingle*) currentListView->currentItem())->item->run(); // doesn't handle "#" in filenames correctly
  }

  currentListView = NULL;
}

void KomparatorWidget::slotOpenFolder()
{
  if ( currentListView == NULL ) return;

  if ( ((KListViewItemSingle*) currentListView->currentItem())->item->isDir() )
  {
    new KRun( ((KListViewItemSingle*) currentListView->currentItem())->item->url() );
    return;
  }

  KURL url = ((KListViewItemSingle*) currentListView->currentItem())->item->url();
  url.setFileName( QString::null );

  new KRun( url );

  currentListView = NULL;
}

void KomparatorWidget::slotCopySelection()
{
  if ( currentListView == NULL ) return;

  // Here we do the same as in QDragObject *KListView::dragObject().
  // this function is protected and therefore not usable.
  KURL::List urls;

  QPixmap *pm = NULL;

  QListViewItemIterator it( currentListView );
  while ( it.current() )
  {
    if ( it.current()->isSelected() )
    {
      urls.append( ((KListViewItemSingle*)it.current())->item->url().url() );
      if ( pm == NULL )
      {
        pm = new QPixmap( ((KListViewItemSingle*)it.current())->item->pixmap(32) );
      }
    }
    it++;
  }

  KURLDrag *ret = new KURLDrag( urls, currentListView );

  if ( pm != NULL && !pm->isNull() ) ret->setPixmap( *pm );

  if ( ret )
  {
    QClipboard *cb = kapp->clipboard();
    cb->setData( ret );
  }

  currentListView = NULL;
}

void KomparatorWidget::slotTrashFiles()
{
  if ( currentListView == NULL ) return;

  if ( currentListView->selectedItems().count() == 0 ) return;

  QString tmp = i18n( "Do you really want to throw the %1 selected file(s) away?" ).arg( currentListView->selectedItems().count() );

  if ( KMessageBox::warningContinueCancel( parentWidget(), tmp, "", KGuiItem( i18n( "Trash" ), "edittrash") ) == KMessageBox::Cancel )
    return;

  QString progress_str = i18n( "Trashing %1 files..." ).arg( currentListView->selectedItems().count() );

  emit( signalProgress( progress_str, -1 ) );

  KURL::List to_delete;

  // Iterate on all selected elements
  QPtrList<QListViewItem> selected = currentListView->selectedItems();
  for ( uint i = 0; i < selected.count(); i++ )
  {
    emit( signalProgress( progress_str, (int)((i*100)/selected.count()) ) );

    if ( currentListView != duplicates1_KListView )
    {
      to_delete.append( ((KListViewItemSingle *)selected.at(i))->item->url() );
      removeFromLists( ((KListViewItemSingle *)selected.at(i))->item );
    }
    else
    {
      to_delete.append( ((KListViewItemDups *)selected.at(i))->item->url() );
      removeFromLists( ((KListViewItemDups *)selected.at(i))->item );
    }
  }

  emit( signalProgress( progress_str, 100 ) );

  KIO::CopyJob *job =  KIO::trash( to_delete );

  if ( findEmptyFolders_CheckBox->isChecked() )
  {
    connect( job, SIGNAL( result (KIO::Job *) ), this, SLOT( slotTrashed (KIO::Job *) ) );
  }

  emit( signalProgress( "", -1 ) );
}

void KomparatorWidget::slotTrashed( KIO::Job *job )
{
  KURL::List urls = ((KIO::CopyJob*)job)->srcURLs();
  checkEmptyDirs( urls );

  disconnect( job, SIGNAL( result (KIO::Job *) ), this, SLOT( slotTrashed (KIO::Job *) ) );

  currentListView = NULL;
}

void KomparatorWidget::slotDeleteFiles()
{
  if ( currentListView == NULL ) return;

  if ( currentListView->selectedItems().count() == 0 ) return;

  QString tmp = i18n( "Do you really want to delete the %1 selected file(s)?" ).arg( currentListView->selectedItems().count() );

  if ( KMessageBox::warningContinueCancel( parentWidget(), tmp, "", KGuiItem( i18n( "Delete" ), "editdelete") ) == KMessageBox::Cancel )
    return;

  QString progress_str = i18n( "Deleting %1 files..." ).arg( currentListView->selectedItems().count() );

  emit( signalProgress( progress_str, -1 ) );

  KURL::List to_delete;

  // Iterate on all selected elements
  QPtrList<QListViewItem> selected = currentListView->selectedItems();
  for ( uint i = 0; i < selected.count(); i++ )
  {
    emit( signalProgress( progress_str, (int)((i*50)/selected.count()) ) );

    if ( currentListView != duplicates1_KListView )
    {
      to_delete.append( ((KListViewItemSingle *)selected.at(i))->item->url() );
      removeFromLists( ((KListViewItemSingle *)selected.at(i))->item );
    }
    else
    {
      to_delete.append( ((KListViewItemDups *)selected.at(i))->item->url() );
      removeFromLists( ((KListViewItemDups *)selected.at(i))->item );
    }
  }

  KURL::List::iterator it;
  int progress = 0;
  for ( it = to_delete.begin(); it != to_delete.end(); it++ )
  {
    emit( signalProgress( progress_str, (int)((progress*50)/to_delete.count())+50 ) );
    KIO::NetAccess::del( (*it), this );
    progress++;
  }

  if ( findEmptyFolders_CheckBox->isChecked() ) checkEmptyDirs( to_delete );

  currentListView = NULL;

  emit( signalProgress( progress_str, 100 ) );

  KIO::CopyJob *job =  KIO::trash( to_delete );

  if ( findEmptyFolders_CheckBox->isChecked() )
  {
    connect( job, SIGNAL( result (KIO::Job *) ), this, SLOT( slotTrashed (KIO::Job *) ) );
  }

  emit( signalProgress( "", -1 ) );
}

void KomparatorWidget::removeFromLists( KFileItemExt *item )
{
  QPtrList< KListViewItemSingle > to_delete_single;
  if ( !item->isDir() )
  {
    // remove from missing list.
    if ( item->dir == 0 ) // left
    {
      QListViewItemIterator it( missing1_KListView );
      while( it.current() )
      {
        if ( ((KListViewItemSingle*)it.current())->item->url() == item->url() )
        {
          to_delete_single.append( (KListViewItemSingle*)it.current() );
        }
        it++;
      }
    }
    else                  // right
    {
      QListViewItemIterator it( missing2_KListView );
      while( it.current() )
      {
        if ( ((KListViewItemSingle*)it.current())->item->url() == item->url() )
        {
          to_delete_single.append( (KListViewItemSingle*)it.current() );
        }
        it++;
      }
    }

    // remove from newer list.
    if ( item->dir == 0 ) // left
    {
      QListViewItemIterator it2( newer1_KListView );
      while( it2.current() )
      {
        if ( ((KListViewItemSingle*)it2.current())->item->url() == item->url() )
        {
          to_delete_single.append( (KListViewItemSingle*)it2.current() );
        }
        it2++;
      }
    }
    else                  // right
    {
      QListViewItemIterator it2( newer2_KListView );
      while( it2.current() )
      {
        if ( ((KListViewItemSingle*)it2.current())->item->url() == item->url() )
        {
          to_delete_single.append( (KListViewItemSingle*)it2.current() );
        }
        it2++;
      }
    }
  }

  QListViewItemIterator it3( duplicates2_KListView );
  while( it3.current() )
  {
    if ( ((KListViewItemSingle*)it3.current())->item->url() == item->url() )
    {
      to_delete_single.append( (KListViewItemSingle*)it3.current() );
    }
    it3++;
  }

  if ( item->brother != NULL ) item = item->brother;

  item->isdupe = 0;  // Make sure the item is removed from duplicates1_KListView, if necessary.
  item->hasdupes = 0;

  // skip this file in duplicate chain
  if ( item->dup_parent != NULL ) // not the first in duplicate chain.
  {
    KFileItemExt *curfile = item->dup_parent;
    curfile->duplicates = item->duplicates;
  }
  else // first in duplicate chain
  {
    if ( item->duplicates != NULL )
    {
      item->duplicates->dup_parent = NULL;
      item->duplicates->hasdupes = 1;
      new KListViewItemDups( duplicates1_KListView, item->duplicates );
    }
    else // Last item in duplicate chain. only for directories.
    {
      if ( item == dirs_dupes ) dirs_dupes = NULL;
    }
  }

  QPtrListIterator< KListViewItemSingle > it4( to_delete_single );
  while( it4.current() )
  {
    delete it4.current();

    ++it4;
  }

  QPtrList< KListViewItemDups > to_delete_dups;
  QListViewItemIterator it5( duplicates1_KListView );
  while( it5.current() )
  {
    if ( ((KListViewItemDups*)it5.current())->item->isdupe == 0 && ((KListViewItemDups*)it5.current())->item->hasdupes == 0 )
    {
      to_delete_dups.append( ((KListViewItemDups*)it5.current()) );
    }
    else if ( ((KListViewItemDups*)it5.current())->item->duplicates == NULL && !((KListViewItemDups*)it5.current())->item->isDir() )
    {
      to_delete_dups.append( ((KListViewItemDups*)it5.current()) );
    }

    it5++;
  }

  QPtrListIterator< KListViewItemDups > it6( to_delete_dups );
  while( it6.current() )
  {
    delete it6.current();

    ++it6;
  }
}

void KomparatorWidget::checkEmptyDirs( KURL::List &urls ) // if a file was deleted, the folder might be empty now.
{
  bool already_appended;
  KURL::List appended_urls;

  KURL::List::iterator it;
  for ( it=urls.begin(); it != urls.end(); it++ )
  {
    if ( QDir( (*it).upURL().path() ).entryList().size() == 2 ) // . and ..
    {
      already_appended = FALSE;

      KURL::List::iterator it2;
      for ( it2=appended_urls.begin(); it2 != appended_urls.end(); it2++ )
      {
        if ( (*it2) == (*it).upURL() ) already_appended = TRUE;
      }

      if ( !already_appended )
      {
        if ( dirs_dupes == NULL ) // no empty dir yet.
        {
          dirs_dupes = new KFileItemExt( (*it).upURL().path( -1 ), "", 0 );
          dirs_dupes->refreshMimeType();
          dirs_dupes->isdupe = 1;
          new KListViewItemDups( duplicates1_KListView, dirs_dupes );
        }
        else // have already empty dir.
        {
          KFileItemExt *item = dirs_dupes->duplicates;
          dirs_dupes->duplicates = new KFileItemExt( (*it).upURL().path( -1 ), "", 0 );
          dirs_dupes->duplicates->refreshMimeType();
          dirs_dupes->hasdupes = 1;
          dirs_dupes->duplicates->hasdupes = 1;
          dirs_dupes->duplicates->isdupe = 1;
          dirs_dupes->duplicates->duplicates = item;
        }

        appended_urls.append( (*it).upURL() );
      }
    }
  }
}

void KomparatorWidget::slotOpenWith()
{
  if ( currentListView != NULL ) KRun::displayOpenWithDialog( KURL::split( ((KListViewItemSingle*) currentListView->currentItem())->item->url() ) );
}

void KomparatorWidget::slotFileProperties()
{
  if ( currentListView != NULL ) new KPropertiesDialog( ((KListViewItemSingle*) currentListView->currentItem())->item, this, "propDialog", TRUE );
}

void KomparatorWidget::slotMissingL2R()
{
  currentListView = missing1_KListView;
  slotCopyToOtherSide();
}

void KomparatorWidget::slotMissingR2L()
{
  currentListView = missing2_KListView;
  slotCopyToOtherSide();
}

void KomparatorWidget::slotNewerL2R()
{
  currentListView = newer1_KListView;
  slotCopyToOtherSide();
}

void KomparatorWidget::slotNewerR2L()
{
  currentListView = newer2_KListView;
  slotCopyToOtherSide();
}

void KomparatorWidget::slotRemoveSelectionMissing1()
{
  currentListView = missing1_KListView;
  slotRemoveFromList();
}

void KomparatorWidget::slotRemoveSelectionMissing2()
{
  currentListView = missing2_KListView;
  slotRemoveFromList();
}

void KomparatorWidget::slotRemoveSelectionNewer1()
{
  currentListView = newer1_KListView;
  slotRemoveFromList();
}

void KomparatorWidget::slotRemoveSelectionNewer2()
{
  currentListView = newer2_KListView;
  slotRemoveFromList();
}

void KomparatorWidget::slotTrashDuplicates()
{
  currentListView = duplicates2_KListView;
  slotTrashFiles();
}

void KomparatorWidget::slotDeleteDuplicates()
{
  currentListView = duplicates2_KListView;
  slotDeleteFiles();
}



/****************************************
 * Internal Functions                   *
 ****************************************/

void KomparatorWidget::slotSizeTypeChanged( int i )
{
  bytes_KIntNumInput->setEnabled( i );
  bytes_KComboBox->setEnabled( i );
}

void KomparatorWidget::slotMD5ChecksumEnabled( bool e )
{
  binaryComparison_CheckBox->setEnabled( e );
  if ( e )
  {
    if ( binaryComparison_CheckBox->isChecked() )
      binaryComparison_KIntSpinBox->setEnabled( TRUE );
    else
      binaryComparison_KIntSpinBox->setEnabled( FALSE );
  }
  else
  {
    binaryComparison_KIntSpinBox->setEnabled( FALSE );
  }
}

void KomparatorWidget::slotStatus( QString str )
{
  status_KLineEdit->setText( str );
}

void KomparatorWidget::slotProgress( QString str, int percent )
{
  if ( percent < 0 || percent > 100 ) progress_KLineEdit->setText( str );
  else progress_KLineEdit->setText( str + i18n( " %1% done." ).arg( percent ) );
}

void KomparatorWidget::slotEmptyFolder( KIO::Job *job, const KIO::UDSEntryList &list )
{
  KIO::UDSEntryListConstIterator end = list.end();
  for ( KIO::UDSEntryListConstIterator it = list.begin(); it != end; ++it )
  {
    if ( KFileItem(*it, "", true, true).url().fileName() != "" && KFileItem(*it, "", true, true).url().fileName() != ".." )
    {
      disconnect( job, SIGNAL( result( KIO::Job * ) ), this, SLOT( slotEmptyFolderFinished( KIO::Job * ) ) );

      job->kill();
      break;
    }
  }
}

void KomparatorWidget::slotEmptyFolderFinished( KIO::Job *job )
{

  KFileItemExt *new_item_dupes = new KFileItemExt( ((KIO::SimpleJob*)job)->url(), "", S_IFDIR );
  new_item_dupes->refreshMimeType();

// FIXME: not very efficient. does not displays modification time or "folder size".

// FIXME: could be searching already in the other url.
  if ( status == SEARCHING_URL_1 )
  {
    new_item_dupes->dir = 0;
  }
  else
  {
    new_item_dupes->dir = 1;
  }

  new_item_dupes->isdupe = 1;
  if ( dirs_dupes != NULL )
  {
    new_item_dupes->hasdupes = 1;
    dirs_dupes->dup_parent = new_item_dupes;
  }

  new_item_dupes->duplicates = dirs_dupes;

  dirs_dupes = new_item_dupes;
}

void KomparatorWidget::enable( bool e )
{
  url1_KURLRequester->setEnabled( e );
  url2_KURLRequester->setEnabled( e );
  searchOptions_ButtonGroup->setEnabled( e );
  search_ButtonGroup->setEnabled( e );
  includeSubfolders_CheckBox->setEnabled( e );
  useFilesIndex_CheckBox->setEnabled( e );
  caseSensitiveSearch_CheckBox->setEnabled( e );
  main_TabWidget->setEnabled( e );
  includeSubfolders_CheckBox->setEnabled( e );
  useFilesIndex_CheckBox->setEnabled( e );
  about_KPushButton->setEnabled( e );
}

void KomparatorWidget::resizeListViewColumns()
{
  int i;
  for ( i=0; i<4; i++ )
  {
    duplicates1_KListView->setColumnWidth( i, duplicates1_width[i] );
    duplicates2_KListView->setColumnWidth( i, duplicates2_width[i] );
    missing1_KListView->setColumnWidth( i, missing1_width[i] );
    missing2_KListView->setColumnWidth( i, missing2_width[i] );
    newer1_KListView->setColumnWidth( i, newer1_width[i] );
    newer2_KListView->setColumnWidth( i, newer2_width[i] );
  }
  duplicates2_KListView->setColumnWidth( 4, duplicates2_width[4] );
}

void KomparatorWidget::slotSaveListViewColumnWidth()
{
  int i;
  for ( i=0; i<4; i++ ) // used for resizing if canceled.
  {
    duplicates1_width[i] = duplicates1_KListView->columnWidth( i );
    duplicates2_width[i] = duplicates2_KListView->columnWidth( i );
    missing1_width[i] = missing1_KListView->columnWidth( i );
    missing2_width[i] = missing2_KListView->columnWidth( i );
    newer1_width[i] = newer1_KListView->columnWidth( i );
    newer2_width[i] = newer2_KListView->columnWidth( i );
  }
  duplicates2_width[4] = duplicates2_KListView->columnWidth( 4 );
}



/****************************************
 * KQuery Functions                     *
 ****************************************/

void KomparatorWidget::slotAddFile( const KFileItem* item, const QString & )
{
  KFileItemExt *new_item_dupes;
  KFileItemExt *new_item_comp;

  if ( item->isFile() && ( followSoftLinks_CheckBox->isChecked() || !item->isLink() ) )
  {                                                              // Real files are added always, (soft)links only if user wants to.
    if ( !item->isReadable() )
    {
      unreadable_files++;
//      KMessageBox::error( this, i18n( "Error reading file %1." ).arg( item->url().url() ),
//                          i18n( "Error while finding files" ) ); // message appeared to often and blocked "cancel" button.
    }
    else
    {
      number_of_files++;

      new_item_dupes = new KFileItemExt( *item );
      new_item_comp = new KFileItemExt( *item );

      if ( status == SEARCHING_URL_1 )
      {
        new_item_dupes->dir = 0;
        new_item_comp->dir = 0;
      }
      else
      {
        new_item_dupes->dir = 1;
        new_item_comp->dir = 1;
      }

      new_item_dupes->next = files_dupes;
      new_item_comp->next = files_comp;

      files_dupes = new_item_dupes;
      files_comp = new_item_comp;

      files_comp->brother = files_dupes;
    }
  }
  else if ( item->isDir() && ( followSoftLinks_CheckBox->isChecked() || !item->isLink() ) && findEmptyFolders_CheckBox->isChecked() )
  {
// FIXME: is that too inefficient???
    if ( item->isLocalFile() ) // true for local dir, too.
    {
      if ( QDir( item->url().path() ).entryList().size() == 2 ) // . and ..
      {
        new_item_dupes = new KFileItemExt( *item );

        if ( status == SEARCHING_URL_1 )
        {
          new_item_dupes->dir = 0;
        }
        else
        {
          new_item_dupes->dir = 1;
        }

        new_item_dupes->isdupe = 1;
        if ( dirs_dupes != NULL )
        {
          new_item_dupes->hasdupes = 1;
          dirs_dupes->dup_parent = new_item_dupes;
        }

        new_item_dupes->duplicates = dirs_dupes;

        dirs_dupes = new_item_dupes;
      }
    }
    else
    {
      KIO::Job *job = KIO::listDir( item->url(), FALSE /*progress*/, TRUE /*hidden*/ );
      connect( job, SIGNAL( entries( KIO::Job *, const KIO::UDSEntryList & ) ), this, SLOT( slotEmptyFolder( KIO::Job *, const KIO::UDSEntryList & ) ) );
      connect( job, SIGNAL( result( KIO::Job * ) ), SLOT( slotEmptyFolderFinished( KIO::Job * ) ) );
    }
  }
}

void KomparatorWidget::slotResult( int errorCode, QString errorString )
{
  if ( errorCode == 0 )
  {
    switch( status )
    {
      case SEARCHING_URL_1:
        scan1_KLed->setColor( QColor( 0, 255, 0 ) );  // Green
        scan2_KLed->setColor( QColor( 255, 0, 0 ) );  // Red
        scan3_KLed->setColor( QColor( 255, 0, 0 ) );

        status = SEARCHING_URL_2;

        emit( signalStatus( i18n( "Searching right URL for files..." ) ) );

        query->setPath( KURL( url2_KURLRequester->url() ) ); // url in url2_KURLRequester is already nice from "initializeQuery".

        query->start();
        break;
      case SEARCHING_URL_2:
        scan1_KLed->setColor( QColor( 0, 255, 0 ) );  // Green
        scan2_KLed->setColor( QColor( 0, 255, 0 ) );
        scan3_KLed->setColor( QColor( 255, 0, 0 ) );  // Red

        status = COMPARING;

        kompare->initialize( files_dupes, files_comp, ignoreEmptyFiles_CheckBox->isChecked(), calculateChecksum_CheckBox->isChecked(),
                             binaryComparison_CheckBox->isChecked(), binaryComparison_KIntSpinBox->value(), number_of_files,
                             url1_KURLRequester->url(), url2_KURLRequester->url() );

        emit( signalStatus( i18n( "Comparing URLs..." ) ) );

        kompare->start();
        break;
      default:
        break;
    }
  }
  else if ( errorCode == KIO::ERR_USER_CANCELED )
  {
    if ( status != CLEAN ) slotSearch();
    return; // cancel is handled in slotSearch()
  }
  else if ( errorCode == KIO::ERR_MALFORMED_URL )
  {
    if ( status == SEARCHING_URL_1 ) KMessageBox::sorry( this, i18n( "Please specify an absolute path in the left URL box." ) );
    else if ( status == SEARCHING_URL_2 ) KMessageBox::sorry( this, i18n( "Please specify an absolute path in the right URL box." ) );
    slotSearch(); // Like clicking cancel button.
    return;
  }
  else
  {
    KMessageBox::sorry( this, errorString );
    slotSearch(); // Like cancel
    return;
  };
}

bool KomparatorWidget::initializeQuery()
{
  int size;

  // only start if we have valid dates
  if ( !isDateValid() ) return FALSE;

  if ( url1_KURLRequester->url().isEmpty() || !KFileItem( url1_KURLRequester->url(), "", S_IFDIR ).isDir() )
  {
    KMessageBox::sorry( this, i18n( "Please specify an absolute path in the left URL box." ) );
    return FALSE;
  }
  else if ( url2_KURLRequester->url().isEmpty() || !KFileItem( url2_KURLRequester->url(), "", S_IFDIR ).isDir() )
  {
    KMessageBox::sorry( this, i18n( "Please specify an absolute path in the right URL box." ) );
    return FALSE;
  }

  QString url1_string = url1_KURLRequester->url().stripWhiteSpace()+"/";
  QString url2_string = url2_KURLRequester->url().stripWhiteSpace()+"/";

  while ( url1_string.right( 2 ) == "//" )
  {
    url1_string = url1_string.left( url1_string.length()-1 );
  }

  while ( url1_string.left( 2 ) == "//" )
  {
    url1_string = url1_string.right( url1_string.length()-1 );
  }

  while ( url2_string.right( 2 ) == "//" )
  {
    url2_string = url2_string.left( url2_string.length()-1 );
  }

  while ( url2_string.left( 2 ) == "//" )
  {
    url2_string = url2_string.right( url2_string.length()-1 );
  }

  url1_KURLRequester->setURL( url1_string );
  url2_KURLRequester->setURL( url2_string );

  KURL url1( url1_string );
  KURL url2( url2_string );

  query->setPath( url1.url() );

  QString regex = regex_KLineEdit->text().isEmpty() ? "*" : regex_KLineEdit->text();
  query->setRegExp( regex, caseSensitiveSearch_CheckBox->isChecked() );

  query->setRecursive( includeSubfolders_CheckBox->isChecked() );

  switch ( bytes_KComboBox->currentItem() )
  {
    case 0:
      size = 1; //one byte
      break;
    case 2:
      size = 1048576; //1M
      break;
    case 1:
    default:
      size=1024; //1k
      break;
  }
  size = bytes_KIntNumInput->value() * size;
  if (size < 0)  // overflow
  {
     if ( KMessageBox::warningYesNo( this, i18n( "Size is too big. Disable size check?" ),
            i18n( "Error" ), i18n( "Disable" ), i18n( "Do Not Disable" ) ) == KMessageBox::Yes )
     {
       size_KComboBox->setCurrentItem( 0 );
       size = -1;
     }
     else
     {
        return FALSE;
     }
  }

  switch ( size_KComboBox->currentItem() )
  {
    case 1:
      query->setSizeRange( size, -1 );
      break;
    case 2:
      query->setSizeRange( -1, size );
      break;
    case 3:
      query->setSizeRange( size, size );
      break;
    default:
      query->setSizeRange( -1, -1 );
  }

  // dates
  QDateTime epoch;
  epoch.setTime_t( 0 );

  // Add date predicate
  if ( !modifiedNoRestriction_RadioButton->isChecked() ) // Modified
  {
    if ( modifiedBetween_RadioButton->isChecked() ) // Between dates
    {
      QDate q1, q2;
      modifiedFrom_KDateCombo->getDate( &q1 );
      modifiedTo_KDateCombo->getDate( &q2 );

      // do not generate negative numbers .. find doesn't handle that
      time_t time1 = epoch.secsTo( q1 );
      time_t time2 = epoch.secsTo( q2.addDays( 1 ) ) - 1; // Include the last day

      query->setTimeRange( time1, time2 );
    }
    else
    {
       time_t cur = time( NULL );
       time_t minutes = cur;

       switch ( modifiedLastTime_KComboBox->currentItem() )
       {
          case 0: // minutes
            minutes = modifiedLastTime_KIntNumInput->value();
            break;
          case 1: // hours
            minutes = 60 * modifiedLastTime_KIntNumInput->value();
            break;
          case 2: // days
            minutes = 60 * 24 * modifiedLastTime_KIntNumInput->value();
            break;
          case 3: // months
            minutes = 60 * 24 * (time_t)( modifiedLastTime_KIntNumInput->value() * 30.41667 );
            break;
          case 4: // years
            minutes = 12 * 60 * 24 * (time_t)( modifiedLastTime_KIntNumInput->value() * 30.41667 );
            break;
          default:
            break;
       }

       query->setTimeRange( cur - minutes * 60, 0 );
    }
  }
  else
    query->setTimeRange( 0, 0 );

  query->setUsername( "" );
  query->setGroupname( "" );

  query->setFileType( 0 );

  query->setMimeType( QString::null );

  //Use locate to speed-up search ?
  query->setUseFileIndex( useFilesIndex_CheckBox->isChecked() );

  return TRUE;
}

bool KomparatorWidget::isDateValid()
{
  // All files
  if ( modifiedNoRestriction_RadioButton->isChecked() ) return TRUE;

  if ( modifiedLastTime_RadioButton->isChecked() )
  {
    if (modifiedLastTime_KIntNumInput->value() > 0 ) return TRUE;

    KMessageBox::sorry( this, i18n( "Unable to search within a period which is less than a minute." ) );
    return FALSE;
  }

  // If we can not parse either of the dates or
  // "from" date is bigger than "to" date return FALSE.
  QDate hi1, hi2;

  QString str;
  if ( ! modifiedFrom_KDateCombo->getDate(&hi1).isValid() ||
       ! modifiedTo_KDateCombo->getDate(&hi2).isValid() )
    str = i18n( "The date is not valid." );
  else if ( hi1 > hi2 )
    str = i18n( "Invalid date range." );
  else if ( QDate::currentDate() < hi1 )
    str = i18n( "Unable to search dates in the future." );

  if ( !str.isNull() )
  {
    KMessageBox::sorry( 0, str );
    return FALSE;
  }

  return TRUE;
}



/****************************************
 * KompareJob Functions                 *
 ****************************************/

void KomparatorWidget::customEvent( QCustomEvent *event )
{
  KListViewItemSingle *new_item_single;
  KListViewItemSingle *new_item_single2;
  QFile file;
  KTempFile *tmpFile;

  switch ( event->type() )
  {
    case RESULT_DUPLICATE:
      new KListViewItemDups( duplicates1_KListView, ((KFileItemExt *) event->data()) );

      break;

    case RESULT_COMP_MISSING:
      if ( ((KFileItemExt *) event->data())->dir == 0 )
      {
        new_item_single = new KListViewItemSingle ( missing1_KListView, ((KFileItemExt *) event->data()), NULL );
        new_item_single->setText( 1, ((KFileItemExt *) event->data())->url().directory(FALSE).replace(
          (KURL(url1_KURLRequester->url())).directory(FALSE)+(KURL(url1_KURLRequester->url())).fileName(), "." ) );
      }
      else
      {
        new_item_single = new KListViewItemSingle( missing2_KListView, ((KFileItemExt *) event->data()), NULL );
        new_item_single->setText( 1, ((KFileItemExt *) event->data())->url().directory(FALSE).replace(
          (KURL(url2_KURLRequester->url())).directory(FALSE)+(KURL(url2_KURLRequester->url())).fileName(), "." ) );
      }

      break;

    case RESULT_COMP_NEWER:
      if ( ((QPair<KFileItemExt *, KFileItemExt *> *) event->data())->first->dir == 0 )
      {
        new_item_single = new KListViewItemSingle( newer1_KListView, ((QPair<KFileItemExt *, KFileItemExt *> *) event->data())->first, NULL );
        new_item_single->setText( 1, ((QPair<KFileItemExt *, KFileItemExt *> *) event->data())->first->url().directory(FALSE).replace(
          (KURL(url1_KURLRequester->url())).directory(FALSE)+(KURL(url1_KURLRequester->url())).fileName(), "." ) );

        new_item_single2 = new KListViewItemSingle( newer2_KListView, ((QPair<KFileItemExt *, KFileItemExt *> *) event->data())->second,
                                                    new_item_single );
        new_item_single2->setText( 1, ((QPair<KFileItemExt *, KFileItemExt *> *) event->data())->second->url().directory(FALSE).replace(
          (KURL(url2_KURLRequester->url())).directory(FALSE)+(KURL(url2_KURLRequester->url())).fileName(), "." ) );
        new_item_single2->color = QColor( 0, 0, 255 );

        new_item_single->brother = new_item_single2;
      }
      else
      {
        new_item_single = new KListViewItemSingle( newer2_KListView, ((QPair<KFileItemExt *, KFileItemExt *> *) event->data())->first,
                                                   NULL );
        new_item_single->setText( 1, ((QPair<KFileItemExt *, KFileItemExt *> *) event->data())->first->url().directory(FALSE).replace(
          (KURL(url2_KURLRequester->url())).directory(FALSE)+(KURL(url2_KURLRequester->url())).fileName(), "." ) );

        new_item_single2 = new KListViewItemSingle( newer1_KListView, ((QPair<KFileItemExt *, KFileItemExt *> *) event->data())->second,
                                                   new_item_single );
        new_item_single2->setText( 1, ((QPair<KFileItemExt *, KFileItemExt *> *) event->data())->second->url().directory(FALSE).replace(
          (KURL(url1_KURLRequester->url())).directory(FALSE)+(KURL(url1_KURLRequester->url())).fileName(), "." ) );
        new_item_single2->color = QColor( 0, 0, 255 );

        new_item_single->brother = new_item_single2;
      }
      delete ((QPair<KFileItemExt *, KFileItemExt *> *) event->data());

      break;

    case RESULT_COMP_DIFFERENT:
      if ( ((QPair<KFileItemExt *, KFileItemExt *> *) event->data())->first->dir == 0 )
      {
        new_item_single = new KListViewItemSingle( newer1_KListView, ((QPair<KFileItemExt *, KFileItemExt *> *) event->data())->first, NULL );
        new_item_single->setText( 1, ((QPair<KFileItemExt *, KFileItemExt *> *) event->data())->first->url().directory(FALSE).replace(
          (KURL(url1_KURLRequester->url())).directory(FALSE)+(KURL(url1_KURLRequester->url())).fileName(), "." ) );
        new_item_single->color = QColor( 255, 0, 0 );

        new_item_single2 = new KListViewItemSingle( newer2_KListView, ((QPair<KFileItemExt *, KFileItemExt *> *) event->data())->second,
                                                    new_item_single );
        new_item_single2->setText( 1, ((QPair<KFileItemExt *, KFileItemExt *> *) event->data())->second->url().directory(FALSE).replace(
          (KURL(url2_KURLRequester->url())).directory(FALSE)+(KURL(url2_KURLRequester->url())).fileName(), "." ) );
        new_item_single2->color = QColor( 255, 0, 0 );

        new_item_single->brother = new_item_single2;
      }
      else
      {
        new_item_single = new KListViewItemSingle( newer2_KListView, ((QPair<KFileItemExt *, KFileItemExt *> *) event->data())->first,
                                                   NULL );
        new_item_single->setText( 1, ((QPair<KFileItemExt *, KFileItemExt *> *) event->data())->first->url().directory(FALSE).replace(
          (KURL(url2_KURLRequester->url())).directory(FALSE)+(KURL(url2_KURLRequester->url())).fileName(), "." ) );
        new_item_single->color = QColor( 255, 0, 0 );

        new_item_single2 = new KListViewItemSingle( newer1_KListView, ((QPair<KFileItemExt *, KFileItemExt *> *) event->data())->second,
                                                   new_item_single );
        new_item_single2->setText( 1, ((QPair<KFileItemExt *, KFileItemExt *> *) event->data())->second->url().directory(FALSE).replace(
          (KURL(url1_KURLRequester->url())).directory(FALSE)+(KURL(url1_KURLRequester->url())).fileName(), "." ) );
        new_item_single2->color = QColor( 255, 0, 0 );

        new_item_single->brother = new_item_single2;
      }
      delete ((QPair<KFileItemExt *, KFileItemExt *> *) event->data());

      break;

    case RESULT_FINISHED:
      if ( findEmptyFolders_CheckBox->isChecked() && dirs_dupes != NULL )
      {
        new KListViewItemDups( duplicates1_KListView, dirs_dupes );
      }

      kompare->wait();

      scan1_KLed->setColor( QColor( 0, 255, 0 ) );  // Green
      scan2_KLed->setColor( QColor( 0, 255, 0 ) );
      scan3_KLed->setColor( QColor( 0, 255, 0 ) );

      enable( TRUE );

      if ( unreadable_files )
      {
        KMessageBox::sorry( this, i18n( "%1 files were unreadable." ).arg( unreadable_files ) );
      }
      unreadable_files = 0;

      status = READY;

      emit( signalStatus( i18n( "Displaying result." ) ) );
      emit( signalProgress( "", -1 ) );

      scan_KPushButton->setText( i18n( "Clear" ) );
      scan_KPushButton->setIconSet( SmallIcon( "clear_left" ) );
      break;

    case DOWNLOAD_FILE:

// FIXME: "QString KTempFile::name() const" SECURITY ISSUE?? consult kdecore documentation.

      tmpFile = new KTempFile;
      requested_download = (KFileItemExt *) event->data();
      ((KFileItemExt *) event->data())->setTmpFile( tmpFile );

      requested_download_job = KIO::file_copy( ((KFileItemExt *) event->data())->url(), tmpFile->name(), -1, TRUE /*overwrite*/ );
      connect( requested_download_job, SIGNAL( result( KIO::Job *) ), this, SLOT( slotFileDownloaded( KIO::Job *) ) );
      break;

    case DELETE_FILE:
      KIO::NetAccess::removeTempFile( ((KTempFile *) event->data())->name() );
      delete ((KTempFile *) event->data());
      requested_download = NULL;

      break;

    case PROGRESS_STATUS:
      emit( signalStatus( *((QString *) event->data()) ) );
      delete ((QString *) event->data());
      break;

    case PROGRESS_PROGRESS:
      emit( signalProgress( *((QPair<QString *, int *> *) event->data())->first, *((QPair<QString *, int *> *) event->data())->second ) );
      delete ((QPair<QString *, int *> *) event->data())->first;
      delete ((QPair<QString *, int *> *) event->data())->second;
      delete ((QPair<QString *, int *> *) event->data());
      break;

    default:
      break;
  }
}

void KomparatorWidget::slotFileDownloaded( KIO::Job *job )
{
  if ( requested_download != NULL )
  {
    if ( job->error() != 0 ) job->showErrorDialog( this );

    requested_download->setDownloadError( job->error() );
  }
  else // no requested_download ??? that means, something is terribly wrong. might happen on "cancel". trying to delete temp file anyway.
  {
    KIO::NetAccess::del( ((KIO::FileCopyJob*)job)->destURL().url(), this );
    KIO::NetAccess::del( ((KIO::FileCopyJob*)job)->destURL().url()+".part", this ); // FIXME: ugly.
  }

  requested_download = NULL;
  requested_download_job = NULL;
}

#include "komparatorwidget.moc"
